﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using FATX_Browser.FATX;

namespace FATX_Browser
{
    public partial class Backup : Form
    {
        System.Threading.Thread t;
        System.Threading.Thread TTimer;
        string filePath;
        bool going = false;
        long sectors;
        FATXDrive ourDrive;
        long currentSector;
        Timer timer;
        public Backup(string FilePath, FATXDrive drive)
        {
            InitializeComponent();
            ourDrive = drive;
            filePath = FilePath;
            IO.HDDFATX hddio = new IO.HDDFATX(drive.IsDump, drive.DumpPath);
            sectors = hddio.GetDriveSize(ourDrive.DeviceID);
            sectors /= 0x200;
            hddio.Close();
            timer = new Timer();
            timer.Tick += new EventHandler(timer_Tick);
            timer.Interval = 150;
            progressBar1.Maximum = (int)(sectors / 2);
            this.FormClosing += new FormClosingEventHandler(Backup_FormClosing);

        }

        void Backup_FormClosing(object sender, FormClosingEventArgs e)
        {
            //CloseStuff();
        }

        void timer_Tick(object sender, EventArgs e)
        {
            label1.Invoke(new call_UpdateLabel(UpdateLabel), new object[] { currentSector });
            progressBar1.Invoke(new call_UpdateProgressBar(UpdateProgressBar), new object[] { currentSector });
        }

        private void button1_Click(object sender, EventArgs e)
        {
            CloseStuff();
        }

        private void CloseStuff()
        {
            if (going)
            {
                if (MessageBox.Show("Are you sure you wish to cancel the backup?", "Confirm", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == DialogResult.Yes)
                {
                    try
                    {
                        t.Abort();
                        TTimer.Abort();
                        timer.Enabled = false;
                        if (MessageBox.Show("Backup aborted.  Delete the backup file?", "Success", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
                        {
                            System.IO.File.Delete(filePath);
                        }
                        this.Close();
                    }
                    catch { }
                }
            }
            else { this.Close(); }
        }

        private void StartThread()
        {
            System.Threading.ThreadStart ts = new System.Threading.ThreadStart(Start);
            TTimer = new System.Threading.Thread(new System.Threading.ThreadStart(timer.Start));
            t = new System.Threading.Thread(ts);
            t.Start();
            TTimer.Start();
            timer.Enabled = true;
            going = true;
        }

        private void Start()
        {
            Write w = new Write(ourDrive);
            //w.BackupDrive(ref currentSector, filePath);
            if (w.BackupDrive(ref currentSector, filePath))
            {
                TTimer.Abort();
                timer.Enabled = false;
                MessageBox.Show("Drive backup complete");
                this.Invoke(new call_FormClose(FormClose));
            }
            else
            {
                MessageBox.Show("Error attempting to backup drive");
                TTimer.Abort();
                timer.Enabled = false;
                this.Invoke(new call_FormClose(FormClose));
            }
        }

        private delegate void call_UpdateLabel(long val);
        private void UpdateLabel(long val)
        {
            decimal percent;
            try
            {
                percent = (decimal)((decimal)(currentSector / 0x200) / (decimal)sectors) * 100;
                if (percent < 0)
                {
                    percent = 0;
                }
            }
            catch { percent = 0; }
            label1.Text = "Currently on Sector 0x" + (val / 0x200).ToString("X") + " of 0x" + sectors.ToString("X") + " -- " + percent.ToString("#") + "%";
        }

        private delegate void call_UpdateProgressBar(long val);
        private void UpdateProgressBar(long val)
        {
            progressBar1.Value = (int)((val / 0x200)/2);
        }

        private delegate void call_FormClose();
        private void FormClose()
        {
            this.Close();
        }

        private void button2_Click(object sender, EventArgs e)
        {
            StartThread();
            button2.Enabled = false;
        }
    }
}
