﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using FATX_Browser.FATX;

namespace FATX_Browser.FATX
{
    class Write
    {
        FATXDrive ourDrive;
        Misc m;
        public Write(FATXDrive drive)
        {
            ourDrive = drive;
            m = new Misc();
        }

        public void WriteFile(string Path)
        {

        }

        public void WriteFile(byte[] buffer)
        {

        }

        public bool NewFolder(string name, ref Folder root)
        {
            if (m.CheckFileName(name))
            {
                object[] ourObject = CheckIfBlocksNeeded(root);
                if ((bool)ourObject[0] == false)
                {
                    EntryData EData = new EntryData();
                    byte[] ourArray = BitConverter.GetBytes(m.FatTimeInt(DateTime.Now));
                    byte[] CD = new byte[] { ourArray[0], ourArray[1] };
                    Array.Reverse(CD);
                    byte[] CT = new byte[] { ourArray[2], ourArray[3] };
                    Array.Reverse(CT);
                    EData.CreationDate = BitConverter.ToInt16(CD, 0);
                    EData.CreationTime = BitConverter.ToInt16(CT, 0);
                    EData.FileName = name;
                    EData.FileNameSize = (byte)name.Length;
                    EData.Flags = 0x10;
                    EData.Size = 0x0;
                    EData.StartingCluster = new FATStuff(root).GetFreeBlocks(1, ((EntryData)ourObject[1]).StartingCluster, 0, false )[0];
                    //If we're using a block that we just created (the current block for the parent folder
                    //has free space
                    if (((EntryData)ourObject[1]).FileName == null)
                    {
                        //Go to the next available entry space
                        EData.EntryOffset = ((EntryData)ourObject[1]).EntryOffset + 0x40;
                    }
                        //We're overwriting a deleted entry - use its entryoffset
                    else { EData.EntryOffset = ((EntryData)ourObject[1]).EntryOffset; }
                    Folder f = new Folder(ourDrive, root.PartInfo);
                    f.EData = EData;
                    CreateNewEntry(EData);
                    WriteFATChain(f.BlocksOccupied,f);
                }
                else
                {
                    EntryData EData = new EntryData();
                    byte[] ourArray = BitConverter.GetBytes(m.FatTimeInt(DateTime.Now));
                    EData.CreationDate = ourArray[0];
                    EData.CreationTime = ourArray[1];
                    EData.FileName = name;
                    EData.FileNameSize = (byte)name.Length;
                    EData.Flags = 0x10;
                    EData.Size = 0x0;
                    EData.StartingCluster = new FATStuff(root).GetFreeBlocks(1, ((EntryData)ourObject[1]).StartingCluster, 0, false)[0];
                    EData.EntryOffset = m.GetBlockOffset(EData.StartingCluster, root);
                    List<uint> xblocks = root.BlocksOccupied.ToList();
                    xblocks.Add(EData.StartingCluster);
                    root.BlocksOccupied = xblocks.ToArray();
                    Folder f = new Folder(ourDrive, root.PartInfo);
                    f.EData = EData;
                    CreateNewEntry(EData);
                    WriteFATChain(f.BlocksOccupied, f);
                }
                return true;
            }
            return false;
        }

        public bool CreateNewEntry(EntryData Edata)
        {
            System.IO.BinaryReader br = ourDrive.GetIO();
            //Set our position so that we can read the entry location
            br.BaseStream.Position = m.DownToNearest200(Edata.EntryOffset);
            byte[] buffer = br.ReadBytes(0x200);
            br.Close();
            //Create our binary writer
            System.IO.BinaryWriter bw = new System.IO.BinaryWriter(new System.IO.MemoryStream(buffer));
            //Set our position to where the entry is
            long EntryOffset = Edata.EntryOffset - m.DownToNearest200(Edata.EntryOffset);
            bw.BaseStream.Position = EntryOffset;
            //Write our entry
            bw.Write(Edata.FileNameSize);
            bw.Write(Edata.Flags);
            bw.Write(Encoding.ASCII.GetBytes(Edata.FileName));
            bw.BaseStream.Position = EntryOffset + 0x2C;
            bw.Write(Edata.StartingCluster);
            bw.Write(Edata.Size);
            bw.Write(Edata.CreationDate);
            bw.Write(Edata.CreationTime);
            //Write the creation date/time again because it will wind up being the
            //same as the modified time
            bw.Write(Edata.CreationDate);
            bw.Write(Edata.CreationTime);
            //Close our writer
            bw.Close();
            //Get our IO
            bw = ourDrive.GetWriterIO();
            bw.BaseStream.Position = m.DownToNearest200(Edata.EntryOffset);
            //Write out our buffer
            bw.Write(buffer);
            return true;
        }

        private bool WriteFATChain(uint[] blocksOccupied, Entry e)
        {
            int i = 0;
            foreach (uint ui in blocksOccupied)
            {
                byte[] Bytes = BitConverter.GetBytes(ui);
                Array.Reverse(Bytes);
                if (i != blocksOccupied.Length - 1)
                {
                    WriteBlock(ui, Bytes, e);
                }
                else
                {
                    if (e.PartInfo.EntrySize == Info.PartitionBit.FATX16)
                    {
                        Bytes = new byte[] { 0xFF, 0xFF };
                    }
                    else { Bytes = new byte[] { 0xFF, 0xFF, 0xFF, 0xFF };}
                    WriteBlock(ui, Bytes, e);
                }
                i++;
            }
            return true;
        }

        private object[] CheckIfBlocksNeeded(Folder f)
        {
            //Create our object array that will hold our Bool and Entry for if
            //we need an open block, and if there's a deleted file

            //Create our entry reader so that we can get a return of entries...
            Entries e = new Entries(f);
            //Get our entries in the last block
            EntryData[] eData = e.GetEntries(f.BlocksOccupied[f.BlocksOccupied.Length - 1]);
            //Files span upon multiple blocks... Here we go to the last block that it occupies
            //(the most recent block created), and check if it has any open entries
            
            bool DeletedEntry = true;
            if (eData.Length < 100)
            {
                EntryData newEntry = new EntryData();
                newEntry.EntryOffset = eData[eData.Length - 1].EntryOffset;
                newEntry.StartingCluster = eData[eData.Length - 1].StartingCluster;
                return new object[] { false, newEntry };
            }
            //Check for deleted entries
            foreach (EntryData E in eData)
            {
                if (E.Flags == (byte)Info.FileFlags.Deleted)
                {
                    DeletedEntry = false;
                    return new object[] { DeletedEntry, E };
                }
            }
            return new object[] { DeletedEntry };
        }

        public bool Delete(File f)
        {
            return true;
        }

        private bool DeleteInternal(File f)
        {
            return true;
        }

        public bool Delete(Folder f, ref long ProgressUpdate, ref long ProgressMax)
        {
            return DeleteInternal(f, ref ProgressUpdate, ref ProgressMax);
        }

        private bool DeleteInternal(Folder f, ref long progressUpdate, ref long progressMax)
        {
            uint[] Occupied = f.BlocksOccupied;
            //Set the max progress
            progressMax = Occupied.Length;
            foreach (File g in f.Files())
            {
                progressMax += (uint)g.BlocksOccupied.Length;
            }
            foreach (Folder g in f.SubFolders())
            {
                progressMax += (uint)g.BlocksOccupied.Length;
            }
            //Set progress update
            progressUpdate = 0;
            //Mark the blocks as empty
            foreach (uint ui in Occupied)
            {
                MarkEntryAsDeleted(f);
                switch (f.PartInfo.EntrySize)
                {
                    case Info.PartitionBit.FATX16:
                        WriteBlock(ui, new byte[] { 0x00, 0x00 }, f);
                        break;
                    case Info.PartitionBit.FATX32:
                        WriteBlock(ui, new byte[] { 0x00, 0x00, 0x00, 0x00 }, f);
                        break;
                }
                progressUpdate++;
            }

            foreach (Folder g in f.SubFolders())
            {
                DeleteInternal(g, ref progressUpdate, ref progressMax);
            }

            foreach (File g in f.Files())
            {
                foreach (uint ui in Occupied)
                {
                    switch (f.PartInfo.EntrySize)
                    {
                        case Info.PartitionBit.FATX16:
                            WriteBlock(ui, new byte[] { 0x00, 0x00 }, f);
                            break;
                        case Info.PartitionBit.FATX32:
                            WriteBlock(ui, new byte[] { 0x00, 0x00, 0x00, 0x00 }, f);
                            break;
                    }
                    progressUpdate++;
                }
            }
            
            return true;
        }

        private bool MarkEntryAsDeleted(Entry e)
        {
            System.IO.BinaryReader br = ourDrive.GetIO();
            long position = m.DownToNearest200(e.EntryOffset);
            //Get our offset in our buffer
            long OffsetInBuffer = e.EntryOffset - position;
            br.BaseStream.Position = position;
            //Read our buffer
            byte[] buffer = br.ReadBytes(0x200);
            //Close our binary reader - not needed now
            br.Close();
            //Create our binary writer
            System.IO.BinaryWriter bw = new System.IO.BinaryWriter(new System.IO.MemoryStream(buffer));
            //Seek to the position in the buffer
            bw.BaseStream.Position = OffsetInBuffer;
            //Write the flag
            bw.Write((byte)Info.FileFlags.Deleted);
            //Close our binary writer
            bw.Close();
            //Re-open our binary writer for the drive
            bw = ourDrive.GetWriterIO();
            //Set our position
            bw.BaseStream.Position = m.DownToNearest200(e.EntryOffset);
            //Write out our buffer
            bw.Write(buffer);
            return true;
        }

        private bool WriteBlock(uint ui, byte[] value, Entry e)
        {
            //Create our binary reader
            System.IO.BinaryReader br = ourDrive.GetIO();
            //Set our position to the block offset in the FAT
            br.BaseStream.Position = m.DownToNearest200(m.BlockToFATOffset(ui, e));
            //Get our position in the buffer
            long OffsetInBuffer = m.BlockToFATOffset(ui, e) - br.BaseStream.Position;
            //Read our buffer
            byte[] buffer = br.ReadBytes(0x200);
            //Close our binary reader - we're done with it for the drive for now
            br.Close();
            //Create our binary writer for writing to the buffer
            System.IO.BinaryWriter bw = new System.IO.BinaryWriter(new System.IO.MemoryStream(buffer));
            //Set our position 
            bw.BaseStream.Position = OffsetInBuffer;
            //Write our block
            bw.Write(value);
            bw.Close();
            //Re-open our binary writer in the drive
            bw = ourDrive.GetWriterIO();
            //Set our position
            bw.BaseStream.Position = m.DownToNearest200(m.BlockToFATOffset(ui, e));
            //Write our buffer
            bw.Write(buffer);
            return true;
        }

        public bool Rename(Entry e, string newName)
        {
            return Rename(e.EntryOffset, newName);
        }


        private bool Rename(long entryOffset, string newName)
        {
            if (m.CheckFileName(newName))
            {
                //Create our binary reader/writer
                System.IO.BinaryReader br = ourDrive.GetIO();
                //Round our offset down to the nearest 0x200 boundary so we can read
                //the entry
                //Set our reader to do the same thing
                br.BaseStream.Position = m.DownToNearest200(entryOffset);
                //Split the difference so we can see how far in to our buffer
                //we need to be
                long offsetInBuffer = entryOffset - (m.DownToNearest200(entryOffset));
                //Create our buffer
                byte[] buffer = br.ReadBytes(0x200);
                //Close our reader - we don't need it any more
                br.Close();
                //Create a new binary reader that reads the buffer in memory
                System.IO.BinaryWriter memWriter = new System.IO.BinaryWriter(new System.IO.MemoryStream(buffer));
                //Set our position to the entry position
                memWriter.BaseStream.Position = offsetInBuffer;
                //Write the file name size
                memWriter.Write((byte)newName.Length);
                //Jump forward one byte and clear the old name
                memWriter.BaseStream.Position++;
                for (int i = 0; i < 0x2A; i++)
                {
                    memWriter.Write((byte)0xFF);
                }
                //Go back to where the file name is
                memWriter.BaseStream.Position = offsetInBuffer + 0x2;
                //Write the new name
                memWriter.Write((byte[])Encoding.ASCII.GetBytes(newName));
                memWriter.Close();
                //Open / set our writer's stuff
                System.IO.BinaryWriter bw = ourDrive.GetWriterIO();
                bw.BaseStream.Position = m.DownToNearest200(entryOffset);
                //Write our edited buffer to the drive
                bw.Write((byte[])buffer);
                bw.Close();
                return true;
            }
            else
            {
                throw new Exception("File name not valid");
            }
        }

        public bool BackupDrive(string savePath)
        {
            long garbage = 0;
            return BackupDrive(ref garbage, savePath);
        }

        public bool BackupDrive(ref long Progress, string savePath)
        {
            System.IO.BinaryWriter fileWriter = null;
            System.IO.BinaryReader driveReader = null;
            try
            {
                long driveSize = ourDrive.DriveSize;
                fileWriter = new System.IO.BinaryWriter(new System.IO.FileStream(savePath, System.IO.FileMode.Create));
                driveReader = ourDrive.GetIO();
                for (long i = 0; i < driveSize; i += 0x4000)
                {
                    driveReader.BaseStream.Position = i;
                    Progress = i;
                    byte[] buffer = driveReader.ReadBytes(0x4000);
                    fileWriter.Write(buffer);
                }
                return true;
            }
            catch { fileWriter.Close(); driveReader.Close(); return false; }
        }
    }
}
