﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using FATX_Browser.FATX;
using System.Runtime.InteropServices;

namespace FATX_Browser
{
    public partial class Backup : Form
    {
        const int MF_BYPOSITION = 0x400;

        [DllImport("User32")]

        private static extern int RemoveMenu(IntPtr hMenu, int nPosition, int wFlags);

        [DllImport("User32")]

        private static extern IntPtr GetSystemMenu(IntPtr hWnd, bool bRevert);

        [DllImport("User32")]

        private static extern int GetMenuItemCount(IntPtr hWnd);

        System.Threading.Thread t;
        System.Threading.Thread TTimer;
        string filePath;
        bool going = false;
        long clusters;
        FATXDrive ourDrive;
        long currentSector;
        Timer timer;
        Info.HDDFATX.Partitions p;
        public Backup(string FilePath, FATXDrive drive, Info.HDDFATX.Partitions partition)
        {
            InitializeComponent();
            IntPtr hMenu = GetSystemMenu(this.Handle, false);
            int menuItemCount = GetMenuItemCount(hMenu);
            RemoveMenu(hMenu, menuItemCount - 1, MF_BYPOSITION);
            ourDrive = drive;
            filePath = FilePath;
            p = partition;
        }

        public Backup(string FilePath, FATXDrive drive)
        {
            InitializeComponent();
            IntPtr hMenu = GetSystemMenu(this.Handle, false);
            int menuItemCount = GetMenuItemCount(hMenu);
            RemoveMenu(hMenu, menuItemCount - 1, MF_BYPOSITION);
            ourDrive = drive;
            filePath = FilePath;
            clusters = ourDrive.DriveSize;
            clusters /= 0x4000;
            timer = new Timer();
            timer.Tick += new EventHandler(timer_Tick);
            timer.Interval = 150;
            progressBar1.Maximum = (int)(clusters / 2);
            this.FormClosing += new FormClosingEventHandler(Backup_FormClosing);

        }

        private void ExtractSS()
        {
            ourDrive.ExtractSS(filePath);
            this.Close();
        }

        private void ExtractJosh()
        {
            ourDrive.ExtractJosh(filePath);
            this.Close();
        }

        void Backup_FormClosing(object sender, FormClosingEventArgs e)
        {
            //CloseStuff();
        }

        void timer_Tick(object sender, EventArgs e)
        {
            label1.Invoke(new call_UpdateLabel(UpdateLabel), new object[] { currentSector });
            progressBar1.Invoke(new call_UpdateProgressBar(UpdateProgressBar), new object[] { currentSector });
        }

        private void button1_Click(object sender, EventArgs e)
        {
            CloseStuff();
        }

        private void CloseStuff()
        {
            if (going)
            {
                if (MessageBox.Show("Are you sure you wish to cancel the backup?", "Confirm", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation) == DialogResult.Yes)
                {
                    try
                    {
                        t.Abort();
                        TTimer.Abort();
                        timer.Enabled = false;
                        if (MessageBox.Show("Backup aborted.  Delete the backup file?", "Success", MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
                        {
                            System.IO.File.Delete(filePath);
                        }
                        this.Close();
                    }
                    catch { this.Close(); }
                }
            }
            else { this.Close(); }
        }

        private void StartThread()
        {
            System.Threading.ThreadStart ts = new System.Threading.ThreadStart(Start);
            TTimer = new System.Threading.Thread(new System.Threading.ThreadStart(timer.Start));
            TTimer.Priority = System.Threading.ThreadPriority.AboveNormal;
            timer.Enabled = true;
            TTimer.Start();
            t = new System.Threading.Thread(ts);
            t.Priority = System.Threading.ThreadPriority.Highest;
            t.Start();
            going = true;
        }

        private void Start()
        {
            if (ourDrive.BackupDrive(ref currentSector, filePath))
            {
                TTimer.Abort();
                timer.Enabled = false;
                System.Media.SystemSounds.Asterisk.Play();
                going = false;
                this.Invoke(new call_FormClose(FormClose));
            }
            else
            {
                MessageBox.Show("Error attempting to backup drive");
                TTimer.Abort();
                timer.Enabled = false;
                going = false;
                this.Invoke(new call_FormClose(FormClose));
            }
        }

        private delegate void call_UpdateLabel(long val);
        private void UpdateLabel(long val)
        {
            decimal percent;
            try
            {
                percent = (decimal)((decimal)(currentSector / 0x4000) / (decimal)clusters) * 100;
                if (percent < 0)
                {
                    percent = 0;
                }
            }
            catch { percent = 0; }
            label1.Text = "Currently on cluster 0x" + (val / 0x4000).ToString("X") + " of 0x" + clusters.ToString("X") + " -- " + percent.ToString("#.00") + "%";
            this.Text = "Backup " + percent.ToString("#.00") + "%";
        }

        private delegate void call_UpdateProgressBar(long val);
        private void UpdateProgressBar(long val)
        {
            progressBar1.Value = (int)((val / 0x4000)/2);
        }

        private delegate void call_FormClose();
        private void FormClose()
        {
            this.Close();
        }

        private void button2_Click(object sender, EventArgs e)
        {
            if (!(p == 0x0))
            {
                if (p == Info.HDDFATX.Partitions.SecuritySector)
                {
                    label1.Text = "Doing work";
                    ExtractSS();
                }
                else if (p == Info.HDDFATX.Partitions.Josh)
                {
                    label1.Text = "Doing work";
                    ExtractJosh();
                }
                else
                {
                    this.Close();
                }
            }
            else
            {
                StartThread();
                button2.Enabled = false;
            }
        }
    }
}
