﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using FATX_Browser.FATX;
using System.Drawing;

namespace FATX_Browser.STFS
{
    class STFSInfo
    {
        FATXDrive xDrive;
        File xFile;
        byte[] Buffer;
        Misc m;
        IOReader IO;
        FATX_Browser.FATX.STFSInfo info;
        public STFSInfo(File f)
        {
            xFile = f;
            xDrive = f.Drive;
            m = new Misc();
            info = new FATX.STFSInfo();
        }

        public FATX.STFSInfo GetInfo()
        {
            info.ConsoleID = ConsoleID();
            info.ContentName = DisplayName();
            info.DeviceID = DeviceID();
            info.Magic = Magic();
            info.TitleID = TitleID();
            info.TitleName = GameName();
            info.ProfileID = ProfileID();
            return info;
        }

        public enum Offsets
        {
            TitleID = 0x360,
            ConsoleID = 0x36c,
            DeviceID = 0x3FD,
            DisplayName = 0x411,
            TitleName = 0x1691,
            ProfileID = 0x371,
            ContentImage = 0x171A,
            TitleImage = 0x571A,
        }

        public byte[] Magic()
        {
            CheckBuffer();
            IOReader io = GetIO();
            byte[] buffer = io.ReadBytes(0x4);
            io.Close();
            return buffer;
        }

        public string GameName()
        {
            CheckBuffer();
            IOReader io = GetIO();
            io.BaseStream.Position = (int)Offsets.TitleName;
            byte[] b = io.ReadBytes(0x80);
            io.Close();
            Buffer = null;
            string ss = "";
            for (int i = 1; i < b.Length; i += 2)
            {
                try
                {
                    if (b[i] == 0x0 && b[i + 1] == 0x0)
                    {
                        break;
                    }
                }
                catch { continue; }
                string temp = Encoding.ASCII.GetString(new byte[] { b[i] });
                ss += temp;
            }
            return ss;
        }

        public uint TitleID()
        {
            CheckBuffer();
            IOReader io = GetIO();
            io.BaseStream.Position = (int)Offsets.TitleID;
            uint ui = io.ReadUInt32(true);
            io.Close();
            return ui;
        }

        public byte[] ConsoleID()
        {
            CheckBuffer();
            IOReader io = GetIO();
            io.BaseStream.Position = (int)Offsets.ConsoleID;
            byte[] buffer = io.ReadBytes(0x5);
            io.Close();
            return buffer;
        }

        public byte[] DeviceID()
        {
            CheckBuffer();
            IOReader io = GetIO();
            io.BaseStream.Position = (int)Offsets.DeviceID;
            byte[] buffer = io.ReadBytes(0x14);
            io.Close();
            return buffer;
        }

        public byte[] ProfileID()
        {
            CheckBuffer();
            IOReader io = GetIO();
            io.BaseStream.Position = (int)Offsets.ProfileID;
            byte[] buffer = io.ReadBytes(0x8);
            io.Close();
            return buffer;
        }

        public string DisplayName()
        {
            CheckBuffer();
            IOReader io = GetIO();
            io.BaseStream.Position = (int)Offsets.DisplayName;
            byte[] b = io.ReadBytes(0x80);
            io.Close();
            string ss = "";
            for (int i = 1; i < b.Length; i+=2)
            {
                try
                {
                    if (b[i] == 0x0 && b[i + 1] == 0x0)
                    {
                        break;
                    }
                }
                catch { continue; }
                string temp = Encoding.ASCII.GetString(new byte[] { b[i] });
                ss += temp;
            }
            return ss;
        }

        private IOReader GetIO()
        {
            try
            {
                IO.Close();
            }
            catch { }
            IO = new IOReader(new System.IO.MemoryStream(Buffer));
            return IO;
        }

        private void CheckBuffer()
        {
            if (Buffer == null)
            {
                //Set our position to the lowest 0x200 byte boundary to where we can read our buffer
                long position = m.GetBlockOffset(xFile.Block, xFile);
                IOReader io = xFile.Drive.GetIO();
                io.BaseStream.Position = position;
                //Read our buffer number one
                Buffer = io.ReadBytes(0x2200);
                //Close our io
                io.Close();
            }
        }

        private uint GuessBlock(long offset)
        {
            return (uint)(m.DownToNearestCluster(offset, xFile.PartInfo.ClusterSize) / xFile.PartInfo.ClusterSize);
        }

        private void Close()
        {
            try
            {
                IO.Close();
            }
            catch { }
        }
    }
}
